/// <reference types="node" />

import { ClientOptions } from 'ws';
import { Data } from 'ws';
import { Handler } from 'mitt';
import { SocketDataRes as SocketDataRes_2 } from '@vue/ws-client-socket';
import { WebSocket as WebSocket_2 } from 'ws';
import { WildcardHandler } from 'mitt';

export declare class ClientSocket {
    options: ClientSocketOptions;
    on: {
        <Key extends keyof SocketOnEvent>(type: Key, handler: Handler<SocketOnEvent[Key]>): void;
        (type: "*", handler: WildcardHandler<SocketOnEvent>): void;
    };
    socket: Socket;
    sockSend: SockSend;
    socketReadyState: {
        0: string;
        1: string;
        2: string;
        3: string;
    };
    protected reactiveData: {
        heartTimer: {
            hasRef: () => boolean;
            refresh: () => NodeJS.Timeout;
            [Symbol.toPrimitive]: () => number;
            ref: () => NodeJS.Timeout;
            unref: () => NodeJS.Timeout;
        } | undefined;
        reconnectNum: number;
        reconnectTimer: {
            hasRef: () => boolean;
            refresh: () => NodeJS.Timeout;
            [Symbol.toPrimitive]: () => number;
            ref: () => NodeJS.Timeout;
            unref: () => NodeJS.Timeout;
        } | undefined;
        isSocketReconnect: boolean;
        isSocketClose: boolean;
    };
    private socketEventMap;
    constructor(options: ClientSocketOptionsCon);
    private clientSend;
    /**
     创建websocket
     */
    createSocket(): Promise<number>;
    /**
     * 等待ws的启动 默认等待3次  并不会启动ws 只会等待ws启动成功
     * @params {number} reconnectMax 该次请求最多等待重启次数
     * 传递该值的情况为
     * 1.已经实例化并运行过createSocket
     * 2.readyState 不为1（ws不在运行中）
     * 3.需等待重启最多reconnectMax次
     * **/
    waitCreate(reconnectMax?: number): Promise<number>;
    private websocketInstance;
    private websocketReconnect;
    private websocketEvent;
    private startHeart;
    private stopHeart;
    close(): void;
}

export declare interface ClientSocketOptions {
    url: string;
    logger: Console;
    socketOptions: {
        wsConfig: ClientOptions | string | undefined;
        reconnectMax: number;
        reconnectTime: number;
        heartCmd: string;
        heartTime: number;
    };
    socketSendOptions: {
        reqInterceptors: ReqInterceptors;
        resInterceptors: ResInterceptors;
    };
}

export declare interface ClientSocketOptionsCon {
    url: ClientSocketOptions['url'];
    logger?: ClientSocketOptions['logger'];
    socketOptions?: Partial<ClientSocketOptions['socketOptions']>;
    socketSendOptions?: Partial<ClientSocketOptions['socketSendOptions']>;
}

export declare type Cmd = string;

export declare interface ReactiveData {
    heartTimer: undefined | NodeJS.Timeout;
    reconnectNum: number;
    reconnectTimer: undefined | NodeJS.Timeout;
    isSocketReconnect: boolean;
    isSocketClose: boolean;
}

export declare type ReqContent = Record<string, any>;

export declare type ReqInterceptors = (req: SendDataReq) => SocketData | Promise<SocketData>;

export declare type RequiredAll<T> = {
    [K in keyof T]-?: RequiredAll<Required<T[K]>>;
};

export declare type ResInterceptors = (res: SocketDataRes) => SendDataRes | Promise<SendDataRes>;

export declare type SendCallback = (data: SendCallbackData) => void;

export declare interface SendCallbackData {
    req: SendDataReq;
    res: SendDataRes;
}

export declare type SendCmdCache = Record<UUID, SendCmdCacheItem>;

export declare interface SendCmdCacheItem {
    req: SendData;
    options: SendOptions;
    resolve: (value: SendCallbackData | PromiseLike<SendCallbackData>) => void;
    reject: (reason?: any) => void;
    status: 'pending' | 'resolved' | 'rejected';
    callback?: SendCallback;
    onTimeout?: NodeJS.Timeout;
}

export declare type SendData = {
    uuid: UUID;
    cmd: Cmd;
    content?: ReqContent;
};

export declare interface SendDataReq extends SendData {
}

export declare interface SendDataRes extends SendData {
    error_msg?: string;
    status?: -1 | 0 | 1;
}

export declare interface SendOptions {
    isOff?: boolean;
    timeout?: number;
    onSendBefore?: (req: SendDataReq) => void;
}

export declare type Socket = WebSocket_2 | WebSocket | undefined;

export declare type SocketData = Data | WebSocketData;

export declare type SocketDataRes = SocketData | MessageEvent<SocketData>;

export declare type SocketOnEvent = {
    'socket-change-state': undefined;
    'socket-open': Event | undefined;
    'socket-error': Error | Event;
    'socket-close': {
        code: number;
        reason?: Buffer;
    } | CloseEvent;
    'socket-message': SocketDataRes_2;
    '*': any;
};

export declare class SockSend {
    options: SockSendOptions;
    sendCmdCache: SendCmdCache;
    on: {
        <Key extends string>(type: Key, handler: Handler<SendCallbackData>): void;
        (type: "*", handler: WildcardHandler<Record<string, SendCallbackData>>): void;
    };
    constructor(options: SockSendOptions);
    /**
     * 注册事件 把事件存储起来 client -> serve  向serve通信
     * @params cmd 指令
     * @params data 配置项
     * @params data.uuid? 任务id
     * @params data.content? 数据
     * @params data.options? 配置项
     * @params data.options.isOff 是否要被主动注销
     * @params data.options.isCacheCmd 是否缓存
     * @params data.options.timeout 超时时间
     * @params callback({req, res, socket}) 回调函数
     * */
    send(cmd: Cmd, data?: SockSendData, callback?: SendCallback): Promise<SendCallbackData>;
    sendError(error_msg: string, uuid?: UUID): void;
    emit(res: SendDataRes): void;
    off(uuid?: string): void;
}

export declare interface SockSendData {
    content?: ReqContent;
    options?: SendOptions;
    uuid?: UUID;
}

export declare type SockSendOptions = {
    url: string;
    clientSend: (req: SendDataReq) => void;
};

export declare type UUID = string;

export declare type WebSocketData = string | ArrayBufferLike | Blob | ArrayBufferView;

export { }
